/*
 * Lexical scanner for the burgomat grammar
 */

#include <ctype.h>
#include <publib.h>
#include <stdlib.h>
#include "lexer.h"

static char lexer_specials[] = "{}();";

/** Create, and initialize a new lexer object. */
struct lexer *lexer_init(char *input, int maxlen)
{
    struct lexer *lexer;

    // Allocate a structure, return 0 on failure
    if (!(lexer = (struct lexer *)malloc(sizeof(struct lexer)))) return 0;

    // Setup the structure
    lexer->input = input;
    lexer->len = maxlen;
    lexer->offs = 0;
    lexer->row = 1;
    lexer->col = 1;

    return lexer;
}

/** Destroy the lexer */
void lexer_destroy(struct lexer *lex)
{
    free(lex);
}

/*
 * Token getting functions. 
 * Most of these are for internal use!
 * Use only lexer_gettoken()!
 */
static struct token *lexer_quanttoken(struct lexer *lex)
{
    int incr = 0;
    char c, *input = lex->input;
    struct token *tok;

    if (!(tok = (struct token *)malloc(sizeof(struct token)))) return 0;

    // Check once again that this is the right type.
    if (!(input[lex->offs] == '+' || input[lex->offs] == '-'))
	return 0;
    
    // Init the token
    tok->type = TOKEN_QUANT;
    tok->val = 0;

    // Okie, the right type, now lex it!
    c = input[lex->offs];
    while (c == input[lex->offs + incr] && (lex->offs + incr) < lex->len) {
	incr++;
	tok->val++;
    }
    
    // Check sign
    if (input[lex->offs] == '-') tok->val = -tok->val;

    // Fix values in lexer
    lex->offs += incr;
    lex->col += incr;
       
    return tok;
}

static struct token *lexer_numtoken(struct lexer *lex)
{
    int incr = 0;
    char c, *input = lex->input;
    struct token *tok;

    if (!(tok = (struct token *)malloc(sizeof(struct token)))) return 0;

    // Check once again that this is the right type.
    if (!(isdigit(input[lex->offs])))
	return 0;
    
    // Init the token
    tok->type = TOKEN_NUMBER;
    tok->val = 0;
    
    c = input[lex->offs];
    while (isdigit(c) && (lex->offs + incr) < lex->len) {
	tok->val *= 10;
	tok->val += (c - '0');
	incr++;
	c = input[lex->offs + incr];
    }
    
    // Fix values in lexer
    lex->offs += incr;
    lex->col += incr;

    return tok;
}
static struct token *lexer_stringtoken(struct lexer *lex)
{
    int incr = 0;
    char *input = lex->input;
    struct token *tok;

    if (!(tok = (struct token *)malloc(sizeof(struct token)))) return 0;

    // Check once again that this is the right type.
    if (!isalpha(input[lex->offs]))
	return 0;
    
    // Init the token
    tok->type = TOKEN_STRING;
    
    // Count length of string
    while (isalpha(input[lex->offs + incr]) && (lex->offs + incr) < lex->len)
	incr++;

    // copy
    tok->string = strndup(&input[lex->offs], incr);

    // Fix values in lexer
    lex->offs += incr;
    lex->col += incr;

    return tok;
}
static struct token *lexer_spectoken(struct lexer *lex)
{
    char *input = lex->input;
    struct token *tok;

    if (!(tok = (struct token *)malloc(sizeof(struct token)))) return 0;

    tok->type = TOKEN_SPECIAL;
    tok->val = input[lex->offs];

    // Fix values in lexer
    lex->offs++;
    lex->col++;

    return tok;
}

struct token *lexer_gettoken(struct lexer *lex)
{
    char *input = lex->input;
    char c;
    int i;

    while (lex->offs < lex->len) {
	// Determine the type of token by looking at the first char.
	c = input[lex->offs];
	
	// Token is quant
	if (c == '+' || c == '-') return lexer_quanttoken(lex);
	// Token is number
	if (isdigit(c)) return lexer_numtoken(lex);
	// Token is string
	if (isalpha(c)) return lexer_stringtoken(lex);
	// Token is special
	for (i = 0; i < strlen(lexer_specials); i++) 
	    if (c == lexer_specials[i]) return lexer_spectoken(lex);

	// Special handling.
	// Token is newline
	if (c == '\n') {
	    lex->col = 0;
	    lex->row++;
	}
	// Token is junk, drop
	lex->offs++;
	lex->col++;
    }
    return 0;
}

/** Free the token thing */
void lexer_freetoken(struct token *tok)
{
    if (tok->type == TOKEN_STRING) free(tok->string);
    free(tok);
}
